//
//  testCompile.cpp
//  codegen
//
//
//  Created by Navneet Potti on 10/25/15.
//  Copyright (c) 2015 Navneet Potti. All rights reserved.
//

#include "somefunc.h"

#include "llvm/ADT/StringRef.h"
#include "llvm/Analysis/ConstantFolding.h"
#include "llvm/Bitcode/ReaderWriter.h"
#include "llvm/ExecutionEngine/ExecutionEngine.h"
#include "llvm/ExecutionEngine/MCJIT.h"
#include "llvm/IR/Argument.h"
#include "llvm/IR/Constants.h"
#include "llvm/IR/DerivedTypes.h"
#include "llvm/IR/GlobalVariable.h"
#include "llvm/IR/InstIterator.h"
#include "llvm/IR/Instructions.h"
#include "llvm/IR/Intrinsics.h"
#include "llvm/IR/LLVMContext.h"
#include "llvm/IR/LegacyPassManager.h"
#include "llvm/IR/Module.h"
#include "llvm/IR/Type.h"
#include "llvm/IR/Verifier.h"
#include "llvm/Support/ErrorOr.h"
#include "llvm/Support/MemoryBuffer.h"
#include "llvm/Support/TargetSelect.h"
#include "llvm/Support/raw_ostream.h"
#include "llvm/Transforms/IPO/PassManagerBuilder.h"
#include "llvm/Transforms/Utils/BasicBlockUtils.h"
#include "llvm/Transforms/Utils/Cloning.h"

#include <cstdio>
#include <iostream>
#include <memory>
#include <string>
#include <utility>

#define DBG   llvm::outs()  << __FILE__ << ":" << __LINE__ << " "
//int *xyz;

using namespace llvm;

class JitCompiler {
  public:
	JitCompiler(const std::string module_file);

	void LoadModule(const std::string& file);

	template <typename FnType>  
	FnType CompileFunc(FnType fn, const std::string& fn_name) { 
		engine_->finalizeObject();
		DBG << "Finalized module is \n" << *module_ << "\n";
		
		auto fn_ptr = engine_->getFunctionAddress(fn_name);
		return reinterpret_cast<decltype(fn)> (fn_ptr);
	}

	void ChangeGlobal();

  private:
	std::unique_ptr<LLVMContext> context_;
	Module *module_;
	std::unique_ptr<ExecutionEngine> engine_;
};


int main() {
	PRINT_STATE;
	xyz = new int (10);
	PRINT_STATE;
	JitCompiler jit("somefunc.bc");
	
	jit.ChangeGlobal();
	auto fn = jit.CompileFunc(&somefunc, "somefunc");
	assert(somefunc() == fn());
	
	delete xyz;
}

void JitCompiler::ChangeGlobal() {
	// ----------------- #0: Do nothing -----------------
	
	// ----------------- #1: UpdateGlobalMapping -----------------
	//auto g = engine_->FindGlobalVariableNamed("xyz");
	//engine_->updateGlobalMapping(g, &xyz);
	//assert(engine_->getGlobalValueAddress("xyz") == (uint64_t) &xyz);
	
	// ----------------- #2: Replace with new global -----------------
	// ------- Ugly hack that works for globals of type T** ----------
	auto g = engine_->FindGlobalVariableNamed("xyz");
	Constant *addr_i = ConstantInt::get(*context_, APInt(64, (uint64_t) xyz));
	auto addr = ConstantExpr::getIntToPtr(
					addr_i, g->getType()->getPointerElementType());

	GlobalVariable *n = new GlobalVariable(
		*module_,
		g->getType()->getPointerElementType(),
		g->isConstant(),
		g->getLinkage(),
		addr,
		g->getName() + "_new");
	g->replaceAllUsesWith(n);
	n->takeName(g);
	g->eraseFromParent();
}

JitCompiler::JitCompiler(const std::string module_file) {
	InitializeNativeTarget();
	InitializeNativeTargetAsmParser();
	InitializeNativeTargetAsmPrinter();

	context_ = std::unique_ptr<LLVMContext> (new LLVMContext());
	LoadModule(module_file);
	//DBG << "Loaded module is \n" << *module_ << "\n";
}

void JitCompiler::LoadModule(const std::string& file) {
	ErrorOr<std::unique_ptr<MemoryBuffer>> f = MemoryBuffer::getFile(file);
	
	ErrorOr<std::unique_ptr<Module>> e_mod = 
		parseBitcodeFile((f.get())->getMemBufferRef(), *context_);
	module_ = e_mod.get().get();

	engine_ = std::unique_ptr<ExecutionEngine> (
				EngineBuilder(std::move(e_mod.get())).create());
}


